import unittest2

from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker
from pyramid.httpexceptions import HTTPForbidden, HTTPFound
from pyramid import testing
from pyramid.threadlocal import get_current_registry
from pyramid.interfaces import IAuthorizationPolicy, IAuthenticationPolicy
from pyramid.authentication import RepozeWho1AuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from webtest import TestApp
from mock import patch

from penelope.core.models.dashboard import User
from penelope.core.models import Base, Project, Customer, Role, Group, CustomerRequest, Application, TimeEntry
from penelope.core import main
from penelope.core.views import PORRequest

settings = {'test': True,
            'sa.dashboard.url': 'sqlite://',
            'sa.dashboard.echo': False,
            'cache.regions' : 'calculate_matrix, template_caching, default_term',
            'cache.type' : 'memory',
            'cache.calculate_matrix.expire' : '0',
            'project_name': 'Penelope',
            'velruse.openid.store':'openid.store.memstore.MemoryStore',
            'velruse.openid.realm':'localhost',}

def dummy_forbidden(request):
    return HTTPForbidden()


from penelope.core.security import views
views.forbidden = dummy_forbidden

class BaseTestCase(unittest2.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.engine = engine_from_config(settings, prefix='sa.dashboard.')
        cls.Session = sessionmaker()

        #populate basic stuff
        Base.metadata.bind = cls.engine
        Base.metadata.create_all(cls.engine)

        import penelope.core.models
        class MockSession(cls.Session):
            def __call__(self):
                return self
        penelope.core.models.DBSession = MockSession(bind=cls.engine)

    def setUp(self):
        from formalchemy.forms import FieldSet
        self.original_validate = FieldSet.validate
        self.original_sync = FieldSet.sync
        FieldSet.validate = lambda x: True
        FieldSet.sync = lambda x: True

        from pyramid_formalchemy.views import ModelView
        self.original_delete = ModelView.delete
        ModelView.delete = lambda x: HTTPFound(location='/')

        connection = self.engine.connect()

        # begin a non-ORM transaction
        self.trans = connection.begin()

        # bind an individual Session to the connection
        self.session = self.Session(bind=connection)
        Base.session = self.Session

        Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)

        #Moneky patch for testing
        from penelope.core.reports import queries
        self.original_qry_active_projects = queries.qry_active_projects
        queries.qry_active_projects = lambda: self.session.query(Project).filter_by(id='not existing id')

        user = User(email=u'u1@rt.com')
        self.session.add(user)

        customer = Customer(name=u'My Customer')
        self.session.add(customer)

        project = Project(name=u'My Project')
        cr = CustomerRequest(name=u"My CR", id="my-project_1")
        project.add_customer_request(cr)

        customer.add_project(project)
        self.session.add(project)

        self.session.commit()

    def tearDown(self):
        # rollback - everything that happened with the
        # Session above (including calls to commit())
        # is rolled back.
        testing.tearDown()
        self.trans.rollback()
        self.session.close()

        from penelope.core.reports import queries
        queries.qry_active_projects = self.original_qry_active_projects

        from formalchemy.forms import FieldSet
        FieldSet.validate = self.original_validate
        FieldSet.sync = self.original_sync

        from pyramid_formalchemy.views import ModelView
        ModelView.delete = self.original_delete


class IntegrationTestBase(BaseTestCase):

    @classmethod
    def setUpClass(cls):
        super(IntegrationTestBase, cls).setUpClass()
        cls.app = main({}, **settings)

    def authenticate_as(self, roles=None, user_id=None):
        if not user_id:
            self.user = self.session.query(User).filter_by(email=u'u1@rt.com').one()
        else:
            self.user = self.session.query(User).get(user_id)
        if roles:
            self.user.roles_in_context = lambda x: set(roles)
        normal_get = self.app.get
        normal_post = self.app.post

        def authenticated_get(*args, **kwargs):
            kwargs['expect_errors'] = True
            kwargs['extra_environ'] = {'repoze.who.identity': {'user': self.user,
                                       'repoze.who.userid': self.user.id}}
            return normal_get(*args, **kwargs)

        def authenticated_post(*args, **kwargs):
            kwargs['expect_errors'] = True
            kwargs['extra_environ'] = {'repoze.who.identity': {'user': self.user,
                                       'repoze.who.userid': self.user.id}}
            return normal_post(*args, **kwargs)

        self.app.get = authenticated_get
        self.app.post = authenticated_post

    def setUp(self):
        self.app = TestApp(self.app)
        self.config = testing.setUp()
        super(IntegrationTestBase, self).setUp()


class TestQuerySecurity(IntegrationTestBase):

    def setUp(self):
        self.config = testing.setUp()
        super(TestQuerySecurity, self).setUp()
        Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)
        import penelope.core.models
        self.session = penelope.core.models.DBSession()

    def tearDown(self):
        testing.tearDown()
        super(TestQuerySecurity, self).tearDown()

    def _add_policy(self, user):
        from penelope.core import security
        authentication_policy = RepozeWho1AuthenticationPolicy(callback=security.rolefinder)
        authorization_policy = ACLAuthorizationPolicy()
        registry = get_current_registry()
        registry.registerUtility(authorization_policy, IAuthorizationPolicy)
        registry.registerUtility(authentication_policy, IAuthenticationPolicy)
        environ = {'repoze.who.identity': {'user': user,
                   'repoze.who.userid': user.id}}

        from penelope.core.security import acl
        registry.registerAdapter(acl.GenericRoles, (acl.Interface, acl.IUser),
                                 acl.IRoleFinder)
        registry.registerAdapter(acl.ProjectRelatedRoles, (acl.IProjectRelated, acl.IUser),
                                 acl.IRoleFinder)
        registry.registerAdapter(acl.CustomerRelatedRoles, (acl.ICustomer, acl.IUser),
                                 acl.IRoleFinder)
        registry.registerAdapter(acl.UserRoles, (acl.IUser, acl.IUser),
                                 acl.IRoleFinder)

        return PORRequest.blank('/foo', environ=environ)

    def test_empty_security_listing(self):
        user = User(email=u'u1@rt.com')
        self.session.add(user)
        project = Project(name=u'My Project 1')
        self.session.add(project)
        self.session.commit()

        query = self.session.query(Project)
        request = self._add_policy(user)
        filtered_projects = request.filter_viewables(query)
        self.assertEqual(tuple(filtered_projects),())

    def test_security_listing(self):
        user = User(email=u'u2@rt.com')
        self.session.add(user)
        project = Project(name=u'My Project 2')
        self.session.add(project)
        self.session.commit()

        query = self.session.query(Project)
        role = Role(name=u'internal_developer')
        user.roles.append(role)
        group = Group()
        group.roles.append(role)
        group.users.append(user)
        project = query.get('my-project-2')
        project.add_group(group)
        self.session.commit()

        request = self._add_policy(user)
        filtered_projects = request.filter_viewables(query)
        self.assertEqual(len(tuple(filtered_projects)), 1)

    def test_security_listing_with_empties(self):
        user = User(email=u'u3@rt.com')
        self.session.add(user)
        project = Project(name=u'My Project 3')
        self.session.add(project)
        self.session.commit()

        query = self.session.query(Project)

        project2 = Project(name=u'My Project 4')
        self.session.add(project2)

        role = Role(name=u'internal_developer')
        user.roles.append(role)
        group = Group()
        group.roles.append(role)
        group.users.append(user)
        project = query.get('my-project-3')
        project.add_group(group)
        self.session.commit()

        request = self._add_policy(user)
        filtered_projects = request.filter_viewables(query)
        self.assertEqual(len(tuple(filtered_projects)), 1)

    def test_security_listing_global_roles(self):
        user = User(email=u'u4@rt.com')
        self.session.add(user)
        project = Project(name=u'My Project 5')
        self.session.add(project)
        self.session.commit()

        query = self.session.query(Project)

        project2 = Project(name=u'My Project 6')
        self.session.add(project2)

        role = Role(name=u'project_manager')
        user.roles.append(role)
        self.session.commit()

        request = self._add_policy(user)
        filtered_projects = request.filter_viewables(query)
        self.assertEqual(len(tuple(filtered_projects)), 2)

    def test_roles_in_context(self):
        user = User(email=u'roles_matrix')
        internal_developer = Role(name=u'internal_developer')
        user.roles.append(internal_developer)

        self.session.add(user)

        project = Project(name=u'p1_for_roles_matrix')
        external_developer = Role(name=u'external_developer')
        a_role = Role(name=u'a role')
        p1_group_2 = Group(project=project)
        p1_group_2.roles.append(a_role)
        p1_group = Group(project=project)
        p1_group.add_user(user)
        p1_group.roles.append(external_developer)

        self.session.add(project)
        self.session.commit()

        from penelope.core.security import acl
        self.assertItemsEqual(acl.GenericRoles(project, user).get_roles(), set(['internal_developer']))
        self.assertItemsEqual(acl.GenericRoles(None, user).get_roles(), set(['internal_developer']))
        self.assertTrue('internal_developer' not in acl.ProjectRelatedRoles(project, user).get_roles())
        self.assertTrue('external_developer' in acl.ProjectRelatedRoles(project, user).get_roles())
        self.assertTrue('owner' in acl.UserRoles(user, user).get_roles())
        self.assertTrue('internal_developer' in acl.UserRoles(user, user).get_roles())

    def test_project_manager_in_context(self):
        user = User(email=u'roles_matrix')
        self.session.add(user)

        project = Project(name=u'p1_for_roles_matrix')
        cr = CustomerRequest(name=u"My CR", id="my-p1_for_roles_matrix")
        project.add_customer_request(cr)
        pm = Role(name=u'project_manager')
        p1_group = Group(project=project)
        p1_group.add_user(user)
        p1_group.roles.append(pm)

        self.session.add(project)
        self.session.commit()

        from penelope.core.security import acl
        self.assertTrue('project_manager' in acl.ProjectRelatedRoles(project, user).get_roles())


TestQuerySecurity.setUpClass()


class SecurityLocalRolesMatrixTest(IntegrationTestBase):

    def setUp(self):
        super(SecurityLocalRolesMatrixTest, self).setUp()

        Base.metadata.drop_all(self.engine)
        Base.metadata.create_all(self.engine)

        import penelope.core.models
        self.session = penelope.core.models.DBSession()

        user = User(email=u'u1@rt.com')
        self.session.add(user)

        customer = Customer(name=u'My Customer')
        self.session.add(customer)

        project = Project(name=u'My Project')
        cr = CustomerRequest(name=u"My CR", id="my-project_1")
        project.add_customer_request(cr)
        customer.add_project(project)
        self.session.add(project)
        self.session.commit()

    def test_customer_view_for_local_roles(self):
        from penelope.core.models import DBSession
        path = '/admin/Customer/my-customer'
        old = Project.active
        Project.active = Project.activated==True

        #internal_developer
        role = Role(name=u'internal_developer')
        user = DBSession.query(User).get(1)
        user.roles.append(role)
        self.session.commit()

        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=403)
        self.assertEqual(403, res.status_int, msg="For %s %s - got %s - should have 403" % (path, user, res.status_int))

        group = Group()
        group.roles.append(role)
        group.users.append(user)
        project = DBSession.query(Project).get('my-project')
        project.add_group(group)
        self.session.commit()

        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=200)
        self.assertEqual(200, res.status_int, msg="For %s %s - got %s - should have 200" % (path, user, res.status_int))

        #external_developer
        group.project_id = None
        role.id = u'external_developer'
        self.session.commit()

        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=403)
        #self.assertEqual(403, res.status_int, msg="For %s %s - got %s - should have 403" % (path, user, res.status_int))

        group.roles.append(role)
        project.add_group(group)
        self.session.commit()
        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=200)
        self.assertEqual(200, res.status_int, msg="For %s %s - got %s - should have 200" % (path, user, res.status_int))

        #customer
        group.project_id = None
        role.id = u'customer'
        self.session.commit()

        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=403)
        #self.assertEqual(403, res.status_int, msg="For %s %s - got %s - should have 403" % (path, user, res.status_int))

        group.roles.append(role)
        project.add_group(group)
        self.session.commit()
        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=200)
        self.assertEqual(200, res.status_int, msg="For %s %s - got %s - should have 200" % (path, user, res.status_int))

        #wrong role
        group.project_id = None
        role.id = u'stupid_name'
        self.session.commit()

        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=403)
        #self.assertEqual(403, res.status_int, msg="For %s %s - got %s - should have 403" % (path, user, res.status_int))

        group.roles.append(role)
        project.add_group(group)
        self.session.commit()
        self.authenticate_as(user_id=1)
        res = self.app.get(path, status=403)
        #self.assertEqual(403, res.status_int, msg="For %s %s - got %s - should have 403" % (path, user, res.status_int))

        Project.active = old


        #
        # test that user cannot see projects listed in dashboard without a 'view' permission
        #

        # authenticated, no local roles
        user = User(email=u'u9@rt.com')
        self.session.add(user)

        project = Project(name=u'My Project 9')
        self.session.add(project)
        self.session.commit()

        self.authenticate_as(user_id=user.id)

        res = self.app.get('/')
        self.assertNotIn(project.name, res.ubody)

        role = Role(name=u'customer')
        group = Group()
        group.roles.append(role)
        group.users.append(user)

        self.session.commit()

        res = self.app.get('/')
        self.assertNotIn(project.name, res.ubody)


SecurityLocalRolesMatrixTest.setUpClass()


class SecurityMatrixTest(IntegrationTestBase):

    def por_security_matrix(self, path, matrix, POST=False):
        for roles, status in matrix:
            self.authenticate_as(roles=roles)
            if POST: res = self.app.post(path, status=status)
            else: res = self.app.get(path, status=status)
            self.assertEqual(status, res.status_int, msg="For %s %s - got %s - should have %s" % (path, roles, res.status_int, status))

    def por_anonymous(self, path, status, POST=False):
        if POST: res = self.app.post(path, status=status)
        else: res = self.app.get(path, status=status)
        self.assertEqual(status, res.status_int, msg="For %s (Anonymous) - got %s - should have %s" % (path, res.status_int, status))

    def test_root(self):
        path = '/'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      200),
                                        (('owner',),              200),
                                        (('customer',),           200),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                      ])

    def test_login_forms(self):
        self.por_anonymous('/login_form', 200)
        self.por_anonymous('/login_success', 302)
        self.por_anonymous('/change_password', 404)
        self.por_anonymous('/change_password?token=', 200)
        self.por_anonymous('/password_reset_form', 200)
        self.por_anonymous('/password_reset', 200, POST=True)
        self.por_anonymous('/logout', 302)

    def test_view_iterations(self):
        path = '/view_iterations'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_manage_iterations(self):
        path = '/manage_iterations'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_generate_iterations(self):
        path = '/generate_iteration'
        self.por_anonymous(path, 403, POST=True)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          302),
                                        (('project_manager',),    302),
                                        (('administrator',),      302),
                                       ], POST=True)

    def test_add_time_entry(self):
        path = '/add_entry'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_my_entries(self):
        path = '/reports/report_my_entries'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_reports_index(self):
        path = '/reports/index'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_reports_all_entries(self):
        path = '/reports/report_all_entries'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_reports_state_change(self):
        path = '/reports/report_state_change'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_reports_custom(self):
        path = '/reports/report_custom'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_user_listing(self):
        path = '/admin/User'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('administrator',),      200),
                                       ])

    def test_user_view(self):
        user_id = self.session.query(User.id).filter_by(email=u'u1@rt.com').one()
        path = '/admin/User/%s' % user_id
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              200),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      200),
                                       ])

    def test_user_tokens(self):
        user_id = self.session.query(User.id).filter_by(email=u'u1@rt.com').one()
        path = '/admin/User/%s/user_tokens' % user_id
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              200),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      200),
                                       ])

    def test_user_edit(self):
        user_id = self.session.query(User.id).filter_by(email=u'u1@rt.com').one()
        path = '/admin/User/%s/edit' % user_id
        self.por_anonymous(path, 403)
        self.por_anonymous(path, 403, POST=True)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              200),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      200),
                                       ])
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              302),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      302),
                                       ], POST=True)

#    def test_user_delete(self):
#        user_id = self.session.query(User.id).filter_by(email=u'u1@rt.com').one()
#        path = '/admin/User/%s/delete' % user_id
#        self.por_anonymous(path, 403, POST=True)
#        self.por_security_matrix(path, [(('authenticated',),      403),
#                                        (('owner',),              403),
#                                        (('customer',),           403),
#                                        (('external_developer',), 403),
#                                        (('internal_developer',), 403),
#                                        (('secretary',),          403),
#                                        (('project_manager',),    403),
#                                       ], POST=True)

    def test_role_listing(self):
        path = '/admin/Role'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      200),
                                       ])

    def test_role_add(self):
        path = '/admin/Role/new'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      200),
                                       ])

    def test_role_edit(self):
        role = Role(name=u'old_role')
        self.session.add(role)
        self.session.commit()
        path = '/admin/Role/old_role/edit'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    403),
                                        (('administrator',),      200),
                                       ])

#    def test_role_delete(self):
#        role = Role(name='to_be_deleted')
#        self.session.add(role)
#        self.session.commit()
#        path = '/admin/Role/to_be_deleted/delete'
#        self.por_anonymous(path, 403, POST=True)
#        self.por_security_matrix(path, [(('authenticated',),      403),
#                                        (('owner',),              403),
#                                        (('customer',),           403),
#                                        (('external_developer',), 403),
#                                        (('internal_developer',), 403),
#                                        (('secretary',),          403),
#                                        (('project_manager',),    403),
#                                        (('administrator',),      302),
#                                       ], POST=True)

    def test_customers_list(self):
        path = '/admin/Customer'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      200),
                                        (('owner',),              200),
                                        (('customer',),           200),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_customers_add(self):
        path = '/admin/Customer/new'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_customer_view(self):
        path = '/admin/Customer/my-customer'
        old = Project.active
        Project.active = Project.activated==True
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           200),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])
        Project.active = old

    def test_customer_metadata(self):
        path = '/admin/Customer/my-customer/metadata'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           200),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_customer_edit(self):
        path = '/admin/Customer/my-customer/edit'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

#    def test_customer_delete(self):
#        customer = Customer(name=u'To Be Deleted Customer')
#        self.session.add(customer)
#        self.session.commit()
#        path = '/admin/Customer/to-be-deleted-customer/delete'
#        self.por_anonymous(path, 403, POST=True)
#        self.por_security_matrix(path, [(('authenticated',),      403),
#                                        (('owner',),              403),
#                                        (('customer',),           403),
#                                        (('external_developer',), 403),
#                                        (('internal_developer',), 403),
#                                        (('secretary',),          403),
#                                        (('project_manager',),    302),
#                                        (('administrator',),      302),
#                                       ], POST=True)

    def test_customer_time_entries(self):
        path = '/admin/Customer/my-customer/customer_time_entries?customer_id=my-customer'
        old = Project.active
        Project.active = Project.activated==True
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])
        Project.active = old

    def test_customer_add_project(self):
        path = '/admin/Customer/my-customer/add_project'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_list(self):
        path = '/admin/Project'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      200),
                                        (('owner',),              200),
                                        (('customer',),           200),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_view(self):
        path = '/admin/Project/my-project'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           302),
                                        (('external_developer',), 302),
                                        (('internal_developer',), 302),
                                        (('secretary',),          302),
                                        (('project_manager',),    302),
                                        (('administrator',),      302),
                                       ])

    def test_project_metadata(self):
        path = '/admin/Project/my-project/metadata'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_edit(self):
        path = '/admin/Project/my-project/edit'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_configuration(self):
        path = '/admin/Project/my-project/configuration'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_add_group(self):
        path = '/admin/Project/my-project/add_group'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_add_app(self):
        path = '/admin/Project/my-project/add_application'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_project_config_apps(self):
        path = '/admin/Project/my-project/applications'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           200),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

#    def test_project_delete(self):
#        prj = Project(name=u'To Be Deleted')
#        self.session.add(prj)
#        self.session.commit()
#        path = '/admin/Project/to-be-deleted/delete'
#        self.por_anonymous(path, 403, POST=True)
#        self.por_security_matrix(path, [(('authenticated',),      403),
#                                        (('owner',),              403),
#                                        (('customer',),           403),
#                                        (('external_developer',), 403),
#                                        (('internal_developer',), 403),
#                                        (('secretary',),          403),
#                                        (('project_manager',),    302),
#                                        (('administrator',),      302),
#                                       ], POST=True)


    def test_project_customer_requests(self):
        path = '/admin/Project/my-project/customer_requests'
        from penelope.core.forms.project import Backlog
        with patch.object(Backlog, 'fetch_done') as fetch_done:
            fetch_done.return_value = {}
            self.por_anonymous(path, 403)
            self.por_security_matrix(path, [(('authenticated',),      403),
                                            (('owner',),              403),
                                            (('customer',),           200),
                                            (('external_developer',), 200),
                                            (('internal_developer',), 200),
                                            (('secretary',),          200),
                                            (('project_manager',),    200),
                                            (('administrator',),      200),
                                        ])

    def test_project_add_customer_requests(self):
        path = '/admin/Project/my-project/add_customer_request'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_customer_request_view(self):
        path = '/admin/CustomerRequest/my-project_1'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('external_developer',), 403),
                                        (('customer',),           200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_customer_request_view_estimations(self):
        path = '/admin/CustomerRequest/my-project_1/estimations'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_customer_request_workflow(self):
        path = '/admin/CustomerRequest/my-project_1/goto_state?state=estimated'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    302),
                                        (('administrator',),      302),
                                       ])

    def test_application_edit(self):
        app = Application(name=u'testing trac')
        prj = self.session.query(Project).get('my-project')
        prj.add_application(app)
        self.session.commit()
        path = '/admin/Application/1/edit'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def add_initial_roles(self):
        for role in (u'internal_developer', u'external_developer', u'customer',
                     u'secretary', u'project_manager'):
            self.session.add(Role(name=role))
        self.session.commit()

    def test_application_with_acl_view(self):
        self.add_initial_roles()
        app = Application(name=u'testing trac')
        prj = self.session.query(Project).get('my-project')
        prj.add_application(app)
        self.session.commit()
        path = '/admin/Application/1'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_trac_with_acl_view(self):
        self.add_initial_roles()
        with patch('penelope.trac.events.add_trac_to_project'):
            app = Application(name=u'testing trac')
            app.application_type = u'trac'
            prj = self.session.query(Project).get('my-project')
            prj.add_application(app)
            self.session.commit()
        path = '/admin/Application/1'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_tp_list(self):
        path = '/admin/TimeEntry'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_tp_add(self):
        path = '/admin/TimeEntry/new'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 200),
                                        (('internal_developer',), 200),
                                        (('secretary',),          200),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_tp_view(self):
        tp = TimeEntry()
        tp.project_id = 'my-project'
        self.session.add(tp)
        self.session.commit()
        path = '/admin/TimeEntry/1'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              200),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 200),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_tp_edit(self):
        tp = TimeEntry(id=1)
        tp.project_id = 'my-project'
        self.session.add(tp)
        self.session.commit()
        path = '/admin/TimeEntry/1/edit'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_tp_edit_in_new(self):
        tp = TimeEntry(id=1)
        tp.workflow_state = u'new'
        tp.project_id = 'my-project'
        self.session.add(tp)
        self.session.commit()
        path = '/admin/TimeEntry/1/edit'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              200),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          403),
                                        (('project_manager',),    200),
                                        (('administrator',),      200),
                                       ])

    def test_tp_workflow(self):
        tp = TimeEntry(id=1)
        tp.project_id = 'my-project'
        self.session.add(tp)
        self.session.commit()
        path = '/admin/TimeEntry/1/goto_state?state=billable'
        self.por_anonymous(path, 403)
        self.por_security_matrix(path, [(('authenticated',),      403),
                                        (('owner',),              403),
                                        (('customer',),           403),
                                        (('external_developer',), 403),
                                        (('internal_developer',), 403),
                                        (('secretary',),          302),
                                        (('project_manager',),    302),
                                        (('administrator',),      302),
                                       ])

#    def test_tp_delete(self):
#        tp = TimeEntry()
#        tp.project_id = 'my-project'
#        self.session.add(tp)
#        self.session.commit()
#        path = '/admin/TimeEntry/1/delete'
#        self.por_anonymous(path, 403, POST=True)
#        self.por_security_matrix(path, [(('authenticated',),      403),
#                                        (('owner',),              302),
#                                        (('customer',),           403),
#                                        (('external_developer',), 403),
#                                        (('internal_developer',), 403),
#                                        (('secretary',),          403),
#                                        (('project_manager',),    302),
#                                        (('administrator',),      302),
#                                       ], POST=True)


SecurityMatrixTest.setUpClass()



# -*- coding: utf-8 -*-
import zope.component.event
import json

from chameleon import PageTemplate
from beaker.cache import cache_region
from zope.interface import alsoProvides
from pyramid.renderers import get_renderer
from pyramid import httpexceptions as exc
from pyramid.security import has_permission
from pyramid.response import Response
from pyramid.exceptions import NotFound
from pyramid.i18n import TranslationStringFactory
from pyramid_formalchemy.utils import TemplateEngine
from pyramid_formalchemy.resources import Models
from pyramid_formalchemy import actions as factions
from pyramid_formalchemy import events
from formalchemy import config
from formalchemy import FieldSet;FieldSet
from formalchemy import Grid;Grid #pyflakes
from formalchemy import forms;forms
from fa.bootstrap.views import ModelView as Base
from fa.bootstrap import actions
from formalchemy import fields
from formalchemy.exceptions import ValidationError
from webhelpers.paginate import Page

from penelope.core.models.dbsession import DBSession
from penelope.core.models import dashboard
from penelope.core import fanstatic_resources, views
from penelope.core.forms import application_views;application_views
from penelope.core.forms import renderers;renderers


config.engine = TemplateEngine()
_ = TranslationStringFactory('penelope')


newcancel = actions.UIButton(
            id='cancel',
            content=actions._('Cancel'),
            permission='edit',
            _class='btn',
            attrs=dict(href="request.fa_parent_url()"),
            )

editcancel = actions.UIButton(
            id='cancel',
            content=actions._('Cancel'),
            permission='edit',
            _class='btn',
            attrs=dict(href="request.fa_url(request.model_name, request.model_id)"),
            )

save_new = actions.UIButton(
        id='save',
        content=_('Save'),
        permission='new',
        _class='btn btn-success',
        attrs=dict(onclick="jQuery(this).parents('form').submit();"),
        )

actions.defaults_actions['new_buttons'][-1] = newcancel
actions.defaults_actions['edit_buttons'][-1] = editcancel
actions.defaults_actions['show_buttons'].pop(1)
actions.defaults_actions['new_buttons'][0] = save_new


class CrudModels(Models, views.DefaultContext):
    def __init__(self, request):
        Models.__init__(self, request)
        views.DefaultContext.__init__(self, request)


class ModelView(Base):
    pager_args = dict(link_attr={'class': ''},
        curpage_attr={'class': 'active'},
        format="~5~")

    def render(self, **kwargs):
        fanstatic_resources.dashboard.need()
        result = super(ModelView, self).render(**kwargs)
        result['main_template'] = get_renderer('penelope.core:skins/main_template.pt').implementation()
        result['main'] = get_renderer('penelope.core.forms:templates/master.pt').implementation()
        return result

    def models(self, **kwargs):
        self.context.request.models = [dashboard.Project, dashboard.User, dashboard.Group, dashboard.Customer]
        return super(ModelView, self).models(**kwargs)

    def update_grid(self, grid):
        pass

    def breadcrumb(self, **args):
        " This we are ignoring from pyramid_formalchemy "
        return actions.Actions()

    def get_page(self, **kwargs):
        """return a ``webhelpers.paginate.Page`` used to display ``Grid``.
        """
        request = self.request
        def get_page_url(page, partial=None):
            url = "%s?page=%s" % (self.request.path, page)
            if partial:
                url += "&partial=1"
            return url
        options = dict(page=int(request.GET.get('page', '1')),
                       url=get_page_url)
        options.update(kwargs)
        if 'collection' not in options:
            query = self.session.query(request.model_class)
            options['collection'] = request.query_factory(request, query)
        collection = list(request.filter_viewables(options.pop('collection')))

        return Page(collection, **options)

    @factions.action()
    def new(self):
        if not hasattr(self.request,'form_action'):
            self.request.form_action = self.request.model_name
        if self.request.method == 'POST':
            return self.create()
        else:
            return super(ModelView, self).new()

    @factions.action('new')
    def create(self):
        request = self.request
        fs = self.get_fieldset(suffix='Add')

        event = events.BeforeRenderEvent(fs.model, self.request, fs=fs)
        alsoProvides(event, events.IBeforeNewRenderEvent)
        zope.component.event.objectEventNotify(event)

        if request.format == 'json' and request.method == 'PUT':
            data = json.load(request.body_file)
        elif request.content_type == 'application/json':
            data = json.load(request.body_file)
        else:
            data = request.POST

        with_prefix = True
        if request.format == 'json':
            with_prefix = bool(request.params.get('with_prefix'))

        fs = fs.bind(data=data, session=self.session, request=request, with_prefix=with_prefix)
        if self.validate(fs):
            fs.sync()
            self.sync(fs)
            self.session.flush()
            if request.format in ('html', 'xhr'):
                if request.is_xhr or request.format == 'xhr':
                    return Response(content_type='text/plain')
                next = request.POST.get('next') or request.fa_url(request.model_name,fs.model.id)
                return exc.HTTPFound(
                    location=next)
            else:
                fs.rebind(fs.model, data=None)
                return self.render(fs=fs)
        return self.render(fs=fs, id=None)

    def delete(self):
        "We shouldn't use default formalchemy delete."
        request = self.request
        request.add_message(u'You cannot remove the object.', type='danger')
        raise exc.HTTPFound(location=request.fa_url(request.model_name, request.model_instance.id))

    def force_delete(self):
        """Forced only by specific models"""
        request = self.request
        record = request.model_instance

        event = events.BeforeDeleteEvent(record, self.request)
        zope.component.event.objectEventNotify(event)

        if record:
            self.session.delete(record)
        else:
            raise NotFound()

        request.add_message(u'Object deleted.', type='success')
        if request.format == 'html':
            if request.is_xhr or request.format == 'xhr':
                return Response(content_type='text/plain')
            return exc.HTTPFound(location=request.fa_parent_url())
        return self.render(id=request.model_id)


    @factions.action('listing')
    def datatable(self, **kwargs):
        """listing page with datatables.net"""
        fanstatic_resources.datatables.need()

        page = self.get_page(**dict(kwargs, items_per_page=999999999))
        fs = self.get_grid()
        fs = fs.bind(instances=page, request=self.request)
        fs.readonly = True

        event = events.BeforeRenderEvent(self.request.model_class(), self.request, fs=fs, page=page)
        alsoProvides(event, events.IBeforeListingRenderEvent)
        zope.component.event.objectEventNotify(event)

        return self.render_grid(fs=fs, id=None)


    def pick_columns(self, fs, columns):
        # filter out unwanted columns

        for field_name in list(fs._render_fields):
            if field_name not in columns:
                del fs._render_fields[field_name]

        # rearrange the OrderedDict

        for field_name in columns:
            fs._render_fields[field_name] = fs._render_fields.pop(field_name)



from penelope.core.forms import project, customer, application, customer_request, timeentry, group, user, role, contract, kanbanboard

def include_forms(config):
    project.configurate(config)
    customer.configurate(config)
    application.configurate(config)
    customer_request.configurate(config)
    timeentry.configurate(config)
    group.configurate(config)
    user.configurate(config)
    role.configurate(config)
    contract.configurate(config)
    kanbanboard.configurate(config)

    config.override_asset(
        to_override="fa.bootstrap:templates/forms/",
        override_with="penelope.core.forms:templates/forms/")


class AttributeField(fields.AttributeField):
    """
    We are monkey patching formalchemy's AttributeField to add
    additional features like dynamic validators for
    unique columns, etc.
    """
    description = None

    def __init__(self, instrumented_attribute, parent):
        super(AttributeField, self).__init__(instrumented_attribute, parent)

        # add validator if column is unique
        if not self.is_collection and not self.is_readonly() and [c for c in self._columns if c.unique]:
            self.validators.append(unique)

fields.AttributeField = AttributeField


def unique(value, field=None):
    """Successful if value is unique"""
    msg = _('Broken ${value}', mapping={'value': value})
    if not field:
        raise ValidationError(msg)

    session = DBSession()

    filters = {}
    filters[field._column_name] = value
    records = session.query(field.model.__class__).filter_by(**filters)

    if records.count() > 0 and records.one() != field.model:
        msg = _("${value} already exists! Field '${field_column_name}' should be unique!",
                mapping={'value': value, 'field_column_name': field._column_name})
        raise ValidationError(msg)
    return value


def security_create(context, request):
    """
    In order to check properly security for Project-related objects,
    we are using this wrapper function.
    """
    project_id = request.params.get('%s--project_id' % request.model_name, '')
    project = request.session_factory().query(dashboard.Project).get(project_id)
    if project:
        request.challenge_item = project
    if has_permission('new', context, request):
        view = ModelView(context, request)
        return view.create()
    else:
        return exc.HTTPForbidden()


@cache_region('template_caching')
def set_template(body):
    return PageTemplate(body)


def cache_template_init(self, id, content="", alt="", permission=None, attrs=None, **rcontext):
    self.id = id
    self.attrs = attrs or {}
    self.permission = permission
    self.rcontext = rcontext
    if 'id' not in self.attrs:
        self.attrs['id'] = repr(id)
    self.update()
    attributes = u';'.join([u'%s %s' % v for v in self.attrs.items()])
    rcontext.update(attrs=self.attrs, attributes=attributes, id=id)
    body = self.body % self.rcontext
    rcontext.update(content=content, alt=alt)
    self.template = set_template(body)

factions.Action.__init__ = cache_template_init



# -*- coding: utf-8 -*-
from fa.bootstrap import actions
from copy import deepcopy

from penelope.core.forms import ModelView, workflow


def configurate(config):
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='',
        attr='show',
        renderer='fa.bootstrap:templates/admin/show.pt',
        model='penelope.core.models.tp.TimeEntry',
        view=TimeEntryModelView)

    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='edit',
        name='edit',
        attr='edit',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.tp.TimeEntry',
        view=TimeEntryModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.tp.TimeEntry',
        view=TimeEntryModelView)

    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='workflow',
        name='goto_state',
        model='penelope.core.models.tp.TimeEntry',
        view=workflow.goto_state)

    config.formalchemy_model_view('admin',
        view='penelope.core.forms.ModelView',
        context='pyramid_formalchemy.resources.ModelListing',
        model='penelope.core.models.tp.TimeEntry',
        renderer='pyramid_formalchemy:templates/admin/listing.pt',
        attr='listing',
        request_method='GET',
        permission='listing')


timeentry_tabs = actions.TabsActions(actions.TabAction("show",
    content="View",
    permission='view',
    attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")), )


class TimeEntryModelView(ModelView):
    actions_categories = ('buttons', 'tabs')
    defaults_actions = deepcopy(actions.defaults_actions)

    def __init__(self, *args, **kwargs):
        super(TimeEntryModelView, self).__init__(*args, **kwargs)
        self.defaults_actions.update(show_tabs=deepcopy(timeentry_tabs))
        wf = workflow.change_workflow(self.context)
        if wf:
            self.defaults_actions['show_tabs'].append(wf)

    def delete(self):
        """
        For TimeEntry we are always forcing to delete.
        No additional validation
        """
        return self.force_delete()




# -*- coding: utf-8 -*-
import logging
import bleach
import colander
import deform
from sunburnt import SolrInterface
from pyramid.view import view_config
from pyramid.url import current_route_url
from penelope.core.models.dashboard import Trac, User
from penelope.core.models import DBSession
from penelope.core.lib.widgets import SearchButton, PorInlineForm
from deform_bootstrap.widget import ChosenMultipleWidget

log = logging.getLogger('penelope')


def searchable_tracs(request):
    all_tracs = DBSession().query(Trac)
    # a list of tracs user can view:
    if not request.has_permission('manage', None):
        user_projects = [g.project for g in request.authenticated_user.groups]
        viewable_tracs = []
        for project in user_projects:
            for trac in project.tracs:
                viewable_tracs.append(trac)
    else:
        viewable_tracs = all_tracs

    query = """SELECT DISTINCT '%(trac)s' AS trac_name, '%(project)s' AS project_name FROM "trac_%(trac)s".permission
 WHERE username IN ('internal_developer', '%(user)s')"""

    queries = []
    for trac in viewable_tracs:
        queries.append(query % {'trac': trac.trac_name,
                                'project': '%s [%s]' % (trac.project.name, trac.project.customer.name),
                                'user': request.authenticated_user.email})
    sql = '\nUNION '.join(queries)
    sql += ';'
    tracs =  DBSession().execute(sql).fetchall()
    return sorted(tracs, key=lambda s: s[1].lower())


class SearchSchema(colander.MappingSchema):

    tracs = colander.SchemaNode(deform.Set(allow_empty=True),
                      widget=ChosenMultipleWidget(placeholder=u'Select tracs',
                                                  style="width:250px"),
                      missing=colander.null,
                      title=u'')

    realms = colander.SchemaNode(deform.Set(allow_empty=True),
               widget=ChosenMultipleWidget(placeholder=u'Select realms',
                                           style="width:250px",
                                           values=[('', ''),
                                                   ('ticket', 'Ticket'),
                                                   ('wiki', 'Wiki'),
                                                   ('changeset','Changeset')]),
               missing=colander.null,
               title=u'')

    authors = colander.SchemaNode(deform.Set(allow_empty=True),
                    widget=ChosenMultipleWidget(placeholder=u'Select author',
                                                 style="width:250px"),
                    missing=colander.null,
                    title=u'')

    searchable = colander.SchemaNode(typ=colander.String(),
                            title=u'',
                            widget = deform.widget.TextInputWidget(
                                            css_class=u'input-xxlarge',
                                            placeholder=u'Searchable text'),)


@view_config(route_name='search', permission='search', renderer='skin')
def search(request):

    schema = SearchSchema().bind(request=request)
    form = PorInlineForm(
                schema,
                formid='search',
                method='GET',
                buttons=[SearchButton(title=u'Search'),]
            )

    tracs = searchable_tracs(request)
    form['tracs'].widget.values = [('', '')] \
                              + [(t.trac_name, t.project_name) for t in tracs]

    users = DBSession.query(User).order_by(User.fullname)
    form['authors'].widget.values = [('', '')] \
                              + [(a.email, a.fullname) for a in users]

    controls = request.GET.items()
    if not controls:
        return {'form': form.render(),
                'results':[]}
    try:
        appstruct = form.validate(controls)
    except deform.ValidationFailure as e:
        return {'form': e.render(),
                'results':[]}

    params = appstruct.copy()
    if not params['tracs']:
        params['tracs'] = [t.trac_name for t in tracs]

    fs = FullTextSearch(request=request, **params)
    results = fs.get_search_results()
    next_url = None
    previous_url = None
    docs = []

    if results:
        docs = [FullTextSearchObject(**doc) for doc in results]
        records_len = results.result.numFound
        if not fs.page_start + fs.page_size >= records_len: # end of set
            next_query = add_param(request, 'page_start', fs.page_start + fs.page_size)
            next_url = current_route_url(request, _query=next_query)

        if not fs.page_start == 0:
            previous_page = fs.page_start - fs.page_size
            if previous_page < 0:
                previous_page = 0
            previous_query = add_param(request, 'page_start', previous_page)
            previous_url = current_route_url(request, _query=previous_query)

    return {'docs': docs,
            'next': next_url,
            'form': form.render(appstruct=appstruct),
            'previous': previous_url,
            'add_param': add_param,
            'results': results}


def add_param(request, key, value):
    base_query = request.params.copy()
    base_query[key] = value
    return base_query


class FullTextSearch(object):

    def __init__(self, request, tracs=None, searchable=None,
                 realms=None, authors=None):
        if not realms:
            realms = []
        if not authors:
            authors = []
        self.request = request
        self.searchable = searchable.split(' ') # always a sequence
        self.viewable_tracs = list(tracs)
        self.realms = list(realms)
        self.authors = list(authors)
        self.solr_endpoint = request.registry.settings.get('por.solr')

    @property
    def page_start(self):
        return int(self.request.params.get('page_start', 0))

    @property
    def page_size(self):
        return 30

    def get_search_results(self,):
        try:
            return self._do_search(sort_by=['-score'])
        except Exception, e:
            log.error("Couldn't perform Full text search, falling back "
                           "to built-in search sources: %s", e)
            return

    def _build_trac_filter(self, si):
        all_tracs = set([a.trac_name for a in DBSession().query(Trac.trac_name)])
        Q = si.query().Q

        def query_include(items):
            if len(items) > 2:
                return Q(project=items.pop()) | query_include(items)
            elif len(items) == 2:
                return Q(project=items.pop()) | Q(project=items.pop())
            elif len(items) == 1:
                return Q(project=items.pop())
            else:
                return ""

        def query_exclude(items):
            if len(items) > 2:
                return ~Q(project=items.pop()) & ~query_exclude(items)
            elif len(items) == 2:
                return ~Q(project=items.pop()) & ~Q(project=items.pop())
            elif len(items) == 1:
                return ~Q(project=items.pop())
            else:
                return ""

        # if viewable_tracs are > (all_tracs / 2)
        # let's exclude
        if len(self.viewable_tracs) > (len(all_tracs) / 2):
            return query_exclude(all_tracs.difference(self.viewable_tracs))
        # in other cases
        # let's use include
        else:
            return query_include(self.viewable_tracs)

    def _build_realm_filter(self, si):

        Q = si.query().Q

        def query(items):
            if len(items) > 2:
                return Q(realm=items.pop()) | query(items)
            elif len(items) == 2:
                return Q(realm=items.pop()) | Q(realm=items.pop())
            elif len(items) == 1:
                return Q(realm=items.pop())
            else:
                return ""
        return query(self.realms)

    def _build_author_filter(self, si):

        Q = si.query().Q

        def query(items):
            if len(items) > 2:
                return Q(author=items.pop()) | query(items)
            elif len(items) == 2:
                return Q(author=items.pop()) | Q(author=items.pop())
            elif len(items) == 1:
                return Q(author=items.pop())
            else:
                return ""
        return query(self.authors)

    def _do_search(self, sort_by=None):
        si = SolrInterface(self.solr_endpoint)

        searchquery = si.Q(*[si.Q(s) for s in self.searchable])
        query = si.query(searchquery).field_limit(score=True)

        realm_query = self._build_realm_filter(si)
        if realm_query:
            query = query.filter(realm_query)

        author_query = self._build_author_filter(si)
        if author_query:
            query = query.filter(author_query)

        trac_query = self._build_trac_filter(si)
        if trac_query:
            query = query.filter(trac_query)

        for field in sort_by or []:
            query = query.sort_by(field)

        query = query.paginate(start=self.page_start, rows=self.page_size)\
                            .highlight('oneline',
                                    **{'simple.pre':'<span class="highlight">',
                                       'snippets': 3,
                                       'fragsize': 600,
                                       'simple.post':'</span>'})

        # boosting - super hacky but sunburnt is not support bq
        options = query.options()
        options['bq'] = ['realm:ticket^999','status:new^100', 'status:assigned^100',
                         'status:reopened^999', 'status:reviewing^100',
                         'status:accepted^100','(*:* -xxx)^999']
        result = query.interface.search(**options)
        return query.transform_result(result, dict)


class FullTextSearchObject(object):
    '''Minimal behaviour class to store documents going to/comping from Solr.
    '''
    def __init__(self, project, realm, id=None, score=None, status=None,
                 title=None, author=None, changed=None, created=None,
                 oneline=None, involved=None, popularity=None, comments=None,
                 parent_id=None, solr_highlights=None , **kwarg):

        if not involved:
            involved = ()
        if not author:
            author = ()
        self.project = project
        self.author = ', '.join(author)
        self.created = created
        self.popularity = popularity
        self.comments = comments
        self.realm = realm
        self.solr_highlights = solr_highlights
        self._title = ''.join(title)
        self._oneline = oneline
        self.id = id
        self.score = score
        self.parent_id = parent_id
        self.status = status

    @property
    def closed(self):
        return self.status == 'closed'

    @property
    def title(self):
        result = []
        if self.solr_highlights:
            result = self.solr_highlights.get('title')
        if not result:
            result = self._title
        text = ''.join(result)
        return bleach.clean(text, ['span'], ['class'], strip=True)

    @property
    def oneline(self):
        result = []
        if self.solr_highlights:
            result = self.solr_highlights.get('oneline')
        if not result:
            result = self._oneline
        text = ''.join(result)
        return bleach.clean(text, ['span'], ['class'], strip=True)

    def href(self):
        if self.realm == 'changeset':
            return "trac/%s/%s/%s/%s" % (self.project, self.realm, self.id, self.parent_id)
        else:
            return "trac/%s/%s/%s" % (self.project, self.realm, self.id)



import zope.component
import feedparser
import datetime, time

from zope.interface import Interface
from pyramid.renderers import get_renderer

from penelope.core.interfaces import IApplicationView
from penelope.core.models.interfaces import ITrac, IApplication, ITracReport, IGoogleDocs
from penelope.core.gdata.views import get_google_document

gsm = zope.component.getGlobalSiteManager()


class UnknownApplicationRenderer(object):
    def __init__(self, application, context, request):
        self.application = application
        self.context = context
        self.request = request

    def render(self):
        result = {}
        result.update(application=self.application,
            context=self.context,
            template=get_renderer('penelope.core.forms:templates/application_generic_view.pt').implementation(),
            request=self.request)
        return result

gsm.registerAdapter(UnknownApplicationRenderer, (IApplication, Interface, Interface), IApplicationView)


class GoogleDocsRenderer(UnknownApplicationRenderer):
    def render(self):
        result = {}
        result.update(application=self.application,
            context=self.context,
            request=self.request)

        response = get_google_document(self.context, self.request, uri=self.application.api_uri)
        result.update(**response)

        def convert_date(node):
            if not node:
                return ''
            else:
                date = node.text
            date = date[:19]
            return datetime.datetime.fromtimestamp(time.mktime(
                        time.strptime(date,'%Y-%m-%dT%H:%M:%S'))).strftime('%Y-%m-%d %H:%M:%S')

        if 'folder' in response.keys():
            template = get_renderer('penelope.core.forms:templates/application_googlefolder_view.pt').implementation()
            result.update(template=template)
            result.update(convert_date=convert_date)

        elif 'document' in response.keys():
            if response['document']:
                doc_url = response['document'].get_html_link().href.replace('/edit', '/preview')
            else:
                doc_url = ''
            template = get_renderer('penelope.core.forms:templates/application_googledocument_view.pt').implementation()
            result.update(template=template, doc_url=doc_url)

        else:
            template = get_renderer('penelope.core.forms:templates/application_googledocument_view.pt').implementation()
            result.update(template=template, doc_url='')

        return result

gsm.registerAdapter(GoogleDocsRenderer, (IGoogleDocs, Interface, Interface), IApplicationView)


class TracRenderer(UnknownApplicationRenderer): pass

gsm.registerAdapter(TracRenderer, (ITrac, Interface, Interface), IApplicationView)

class TracReportRenderer(UnknownApplicationRenderer):
    def render(self):
        result = {}
        cookie = self.request.headers.get('Cookie')
        url = self.application.application_uri(request=self.request)
        rss_url = '%s%sasc=1&format=rss' % (url, '?' in url and '&' or '?')
        feed = feedparser.parse(rss_url, request_headers={'Cookie': cookie})
        result.update(application=self.application,
            context=self.context,
            feed=feed,
            template=get_renderer('penelope.core.forms:templates/application_trac_report.pt').implementation(),
            request=self.request)
        return result

gsm.registerAdapter(TracReportRenderer, (ITracReport, Interface, Interface), IApplicationView)



# -*- coding: utf-8 -*-

from zope.interface import implements

from penelope.core import fanstatic_resources
from penelope.core.interfaces import IReportView
from penelope.core.views import DefaultContext


class ReportContext(DefaultContext):
    """
    Default context factory for Report views.
    """

    implements(IReportView)

    def __init__(self, request):
        super(ReportContext, self).__init__(request)
        fanstatic_resources.project_filter.need()
        fanstatic_resources.saved_queries.need()



# -*- coding: utf-8 -*-

import os.path, time
import zope.component
from zope.interface import implements
from pyramid.view import view_config
from pyramid.renderers import get_renderer
from pyramid.threadlocal import get_current_registry
from penelope.core.sidebar import SidebarRenderer, HeaderSidebarAction, SidebarAction
from penelope.core.interfaces import ISidebar
from penelope.core.interfaces import IManageView
from penelope.core.views import DefaultContext


class ManageContext(DefaultContext):
    "Default context factory for Manage views."
    implements(IManageView)


class ManageSidebarRenderer(SidebarRenderer):

    def render(self, request):
        self.actions.append(HeaderSidebarAction('manage',content=u'Manage penelope', permission='manage', no_link=True))
        self.actions.append(SidebarAction('manage_users',
                                          content=u'Manage users',
                                          permission='manage',
                                          attrs=dict(href="'%s/admin/User' % request.application_url")))
        self.actions.append(SidebarAction('manage_roles',
                                          content=u'Manage roles',
                                          permission='manage',
                                          attrs=dict(href="'%s/admin/Role' % request.application_url")))
        self.actions.append(SidebarAction('manage_groups',
                                          content=u'Manage groups',
                                          permission='manage',
                                          attrs=dict(href="'%s/admin/Group' % request.application_url")))
        self.actions.append(SidebarAction('manage_svn_authz',
                                          content=u'Manage SVN authz',
                                          permission='manage_svn',
                                          attrs=dict(href="'%s/manage/svn_authz' % request.application_url")))
        actions = self.actions.render(request)
        template =  get_renderer('penelope.core.forms:templates/project_sidebar.pt').implementation()
        return template(actions=actions,
                        request=request)

gsm = zope.component.getGlobalSiteManager()
gsm.registerAdapter(ManageSidebarRenderer, (IManageView,), ISidebar)


@view_config(route_name='administrator', permission='manage', renderer='skin')
def manage_home(request):
    return {}

@view_config(route_name='manage_svn_authz', renderer='skin', permission='manage_svn')
def manage_svn_authz(request):
    settings = get_current_registry().settings
    authz_file = settings.get('por.svn.authz')

    if request.method == 'POST':
        from penelope.trac.authz import generate_authz
        generate_authz(request.registry.settings)

    with open(authz_file, 'r') as configfile:
        authz = configfile.read()

    modified = time.ctime(os.path.getmtime(authz_file))
    return {'authz': authz,
            'authz_modified': modified}



import csv
import os
import StringIO
import tempfile
import xlwt


class CSVReportRenderer(object):
    def __init__(self, info):
        pass

    def __call__(self, value, system):
        fout = StringIO.StringIO()
        writer = csv.writer(fout, delimiter=';', quoting=csv.QUOTE_ALL)

        writer.writerow(value['header'])
        writer.writerows(value['rows'])

        resp = system['request'].response
        resp.content_type = 'text/csv'
        resp.content_disposition = 'attachment;filename="report.csv"'
        return fout.getvalue()


class XLSReportRenderer(object):
    def __init__(self, info):
        pass

    def __call__(self, value, system):
        wb = xlwt.Workbook()
        ws = wb.add_sheet('report')

        style_header = xlwt.XFStyle()
        style_header.font = xlwt.Font()
        style_header.font.bold = True

        header = value['header']

        maxwidth = [0] * len(header)

        for idx, row in enumerate(value['rows']):
            for colnum, value in enumerate(row):
                ws.write(idx + 1, colnum, value)
                # crude attempt to get a sensible width
                maxwidth[colnum] = max(maxwidth[colnum], int((1 + len(unicode(value))) * 256))

        for colnum, colname in enumerate(header):
            ws.write(0, colnum, colname, style=style_header)
            ws.col(colnum).width = maxwidth[colnum]

        resp = system['request'].response
        resp.content_type = 'application/vnd.ms-excel'
        resp.content_disposition = 'attachment;filename="report.xls"'
        return self.wb_as_bytes(wb)


    @staticmethod
    def wb_as_bytes(wb):
        pathname = tempfile.mktemp()
        wb.save(pathname)
        ret = open(pathname, 'rb').read()
        os.unlink(pathname)
        return ret

